superstrict


import brl.linkedlist
import brl.filesystem
import brl.max2d
import brl.pngloader
import brl.timer
import brl.random




type TGamefield
	const ORIGINAL_TILEWIDTH:int = 301
	field width:int
	field height:int
	field scaling:float
	field wrap_enabled:int
	field tiles:TTile[,]
	field tile_width:int
	field lights:TList
	field lamps:TList
	field offsetx:int
	field offsety:int
	function create_empty:TGamefield(width:int,height:int)
		local g:TGamefield = new TGamefield
		g.width = width
		g.height = height
		g.tiles = new TTile[width,height]
		for local x:int = 0 until width
			for local y:int = 0 until height
				g.tiles[x,y] = TEmptyTile.create(x,y)
			next
		next
		return g
	endfunction
	function load_from_file:TGamefield(file:string)
		local g:TGamefield = new TGamefield
		local stream:TStream = readfile(file)
		local head:string[] = stream.readline().split(",")
		g.width = int(head[0])
		g.height = int(head[1])
		g.wrap_enabled = int(head[2])
		g.tiles = new TTile[g.width,g.height]
		
		local column:int = 0
		local row:int = 0
		while not stream.eof()
			column = 0
			local line:string = stream.readline()
			local tiles:string[] = line.split(";")
			for local tile:string = eachin tiles
				local loaded_tile:TTile
				local bits:string[] = tile.split("|")
				select bits[0]
				case "*"
					local color:int = int(bits[1])
					local directions:int[] = strarr_to_intarr(bits[2].split(","))
					loaded_tile = TLight.create(column,row,color,directions)
					g.lights.addlast(loaded_tile)
				case "l"
					local color:int = int(bits[1])
					local rotation:int = int(bits[2])
					loaded_tile = TLamp.create(column,row,color,rotation)
					g.lamps.addlast(loaded_tile)
				case "p"
					local tracekinds:int[] = strarr_to_intarr(bits[1].split(","))
					local rotations:int[] = strarr_to_intarr(bits[2].split(","))
					loaded_tile = TPath.create(column,row,tracekinds,rotations)
				case "e"
					loaded_tile = TEmptyTile.create(column,row)
				endselect
				g.tiles[column,row] = loaded_tile
				column :+ 1
			next
			row :+ 1
		wend
		stream.close()
		return g
	endfunction
	method new()
		lights = new TList
		lamps = new TList
		tile_width = ORIGINAL_TILEWIDTH
		scaling = 1
	endmethod
	method scramble()
		for local i:TTile = eachin tiles
			i.rotate(rand(0,3))
		next
	endmethod
	method calculate_scaling(xavail:int,yavail:int)
		tile_width = min((xavail-5)/width,(yavail-5)/height)
		scaling = float(tile_width) / ORIGINAL_TILEWIDTH
	endmethod
	method run()
		for local i:TTile = eachin tiles
			i.reset_color()
		next
		for local i:TLight = eachin lights
			i.turnon()
		next
	endmethod
	method finished:int()
		for local i:TLamp = eachin lamps
			if not i.is_on() then return false
		next
		return true
	endmethod
	method draw(x:int=-1,y:int=-1)
		if x <> -1 then offsetx = x
		if y <> -1 then offsety = y
		calculate_scaling(graphicswidth()-offsetx,graphicsheight()-offsety)
		local sx:float,sy:float
		getscale(sx,sy)
		setscale scaling,scaling
		for local i:int = 0 until width
			for local n:int = 0 until height
				tiles[i,n].draw((i+0.5)*tile_width+offsetx,(n+0.5)*tile_width+offsety)
			next
		next
		setscale sx,sy
	endmethod
	method neighbour_for_tile:TTile(x:int,y:int,direction:int)
		select direction
		case DIRECTION_UP
			if y = 0 then
				if wrap_enabled then return tiles[x,height-1] else return null
			endif
			return tiles[x,y-1]
		case DIRECTION_RIGHT
			if x = width-1 then
				if wrap_enabled then return tiles[0,y] else return null
			endif
			return tiles[x+1,y]
		case DIRECTION_DOWN
			if y = height-1 then
				if wrap_enabled then return tiles[x,0] else return null
			endif
			return tiles[x,y+1]
		case DIRECTION_LEFT
			if x = 0 then
				if wrap_enabled then return tiles[width-1,y] else return null
			endif
			return tiles[x-1,y]
		endselect
	endmethod
	method mouse(button:int,x:int,y:int)
		local tile:TTile = tile_for_coordinates(x,y)
		if not tile then return
		if button = 1 then
			tile.rotate(1)
		elseif button = 2 then
			tile.rotate(-1)
		endif
	endmethod
	method tile_for_coordinates:TTile(x:int var,y:int var)
		if x < offsetx or y < offsety then return null
		x = (x-offsetx) / tile_width
		y = (y-offsety) / tile_width
		if x >= 0 and x < width and y>= 0 and y < height then return tiles[x,y]
		return null
	endmethod
	method save_to_file(file:string)
		local stream:TStream = writefile(file)
		stream.writeline(width+","+height+","+wrap_enabled)
		
		for local y:int = 0 until height
			local line:string = ""
			for local x:int = 0 until width
				line :+ tiles[x,y].save()
				if x < width-1 then line :+ ";"
			next
			stream.writeline(line)
		next
		
		stream.close()
	endmethod
endtype




type TTile
	field x:int
	field y:int
	field rotation:int
	method rotate(num:int)
		rotation :+ num
		rotation = wrap_var(rotation)
	endmethod
	method rotate_absolute(direction:int)
		local delta:int = direction - rotation
		rotate(delta)
	endmethod
	method is_connected:int(direction:int) abstract
	method add_color(color:int,side:int) abstract
	method reset_color() abstract
	method draw(x:int,y:int) abstract
	method save:string() abstract
	method get_neighbour:TTile(direction:int)
		return gamefield.neighbour_for_tile(x,y,direction)
	endmethod
endtype




type TEmptyTile extends TTile
	function create:TEmptyTile(x:int,y:int)
		local t:TEmptyTile = new TEmptyTile
		t.x = x
		t.y = y
		return t
	endfunction
	method is_connected:int(direction:int)
		return false
	endmethod
	method add_color(color:int,side:int)
	endmethod
	method reset_color()
	endmethod
	method draw(x:int,y:int)
		drawimage img_empty,x,y
	endmethod
	method save:string()
		return "e"
	endmethod
endtype




type TLight extends TTile
	field colors:int[]
	field light_color:int
	field sides:int[]
	function create:TLight(x:int,y:int,color:int,sides:int[])
		local l:TLight = new  TLight
		l.x = x
		l.y = y
		l.set_color(color)
		l.sides = sides
		l.colors = new int[4]
		return l
	endfunction
	method rotate(num:int)
		super.rotate(num)
		for local i:int = 0 until sides.length
			sides[i] :+ num
			sides[i] = wrap_var(sides[i])
		next
		if num > 0 then
			for local i:int = 1 to num
				colors = [colors[3]] + colors[0..3]
			next
		elseif num < 0 then
			for local i:int = num to -1
				colors = colors[1..4] + [colors[0]]
			next
		endif
	endmethod
	method add_side(side:int)
		sides :+ [side]
		colors[side] = COLOR_UNDEFINED
	endmethod
	method set_color(color:int)
		light_color = color
	endmethod
	method is_connected:int(direction:int)
		for local i:int = eachin sides
			if direction = i then return true
		next
		return false
	endmethod
	method add_color(color:int,side:int)
		if is_connected(side) then colors[side] = mix_colors(colors[side],color)
	endmethod
	method reset_color()
		for local i:int = eachin sides
			colors[i] = COLOR_UNDEFINED
		next
	endmethod
	method turnon()
		for local i:int = eachin sides
			colors[i] = mix_colors(colors[i],light_color)
			local n:TTile = get_neighbour(i)
			if n <> null then n.add_color(light_color,opposite_direction(i))
		next
	endmethod
	method draw(x:int,y:int)
		local r:int,g:int,b:int,rot:int=getrotation()
		getcolor(r,g,b)
		for local i:int = eachin sides
			setconstantcolor(colors[i])
			setrotation i*90
			drawimage img_wire,x,y
		next
		setconstantcolor(light_color)
		drawimage img_light,x,y
		setcolor r,g,b
		setrotation rot
	endmethod
	method save:string()
		return "*|"+light_color+"|"+",".join(intarr_to_strarr(sides))
	endmethod
endtype




type TLamp extends TTile
	field dest_color:int
	field color:int
	field side:int
	function create:TLamp(x:int,y:int,color:int,rotation:int)
		local l:TLamp = new TLamp
		l.x = x
		l.y = y
		l.set_color(color)
		l.side = DIRECTION_UP
		l.rotate(rotation)
		return l
	endfunction
	method rotate(num:int)
		super.rotate(num)
		side :+ num
		side = wrap_var(side)
	endmethod
	method set_color(color:int)
		dest_color = color
	endmethod
	method add_color(color:int, side:int)
		if is_connected(side) then
			self.color = mix_colors(self.color,color)
		endif
	endmethod
	method reset_color()
		color = COLOR_UNDEFINED
	endmethod
	method is_on:int()
		return color = dest_color
	endmethod
	method is_connected:int(direction:int)
		return direction = side
	endmethod
	method draw(x:int,y:int)
		local r:int,g:int,b:int,rot:int=getrotation()
		getcolor(r,g,b)
		setrotation rotation*90
		setconstantcolor(color)
		drawimage img_wire,x,y
		setconstantcolor(dest_color,not is_on())
		drawimage img_lamp[is_on()],x,y
		setcolor r,g,b
		setrotation rot
	endmethod
	method save:string()
		return "l|"+dest_color+"|"+rotation
	endmethod
endtype



type TPath extends TTile
	field traces:TList
	function create:TPath(x:int,y:int,traces:int[],rotations:int[])
		local p:TPath = new TPath
		p.x = x
		p.y = y
		p.traces = new TList
		for local i:int = 0 until traces.length
			p.add_trace(traces[i],rotations[i])
		next
		return p
	endfunction
	method add_trace(trace:int,rotation:int)
		rotate(-rotation)
		traces.addlast(TTrace.create(trace,self))
		rotate(rotation)
	endmethod
	method rotate(num:int)
		super.rotate(num)
		for local i:TTrace = eachin traces
			i.rotate(num)
		next
	endmethod
	method is_connected:int(direction:int)
		return find_trace_for_side(direction) <> null
	endmethod
	method find_trace_for_side:TTrace(side:int)
		for local i:TTrace = eachin traces
			if i.is_connected(side) then return i
		next
		return null
	endmethod
	method add_color(color:int,side:int)
		local trace:TTrace = find_trace_for_side(side)
		if trace = null then return
		trace.add_color(color,side)
	endmethod
	method reset_color()
		for local i:TTrace = eachin traces
			i.color = COLOR_UNDEFINED
		next
	endmethod
	method draw(x:int,y:int)
		local r:int,g:int,b:int,rot:int=getrotation()
		getcolor(r,g,b)
		
		setrotation rotation*90
		for local i:TTrace = eachin traces
			i.draw(x,y)
		next
		
		setcolor r,g,b
		setrotation(rot)
	endmethod
	method save:string()
		local kinds:int[traces.count()]
		local rotations:int[traces.count()]
		local nr:int = 0
		for local i:TTrace = eachin traces
			kinds[nr] = i.kind
			rotations[nr] = i.rotation
			nr :+ 1
		next
		return "p|"+",".join(intarr_to_strarr(kinds))+"|"+",".join(intarr_to_strarr(rotations))
	endmethod
endtype



type TTrace
	global tracemaps:int[][] = [[DIRECTION_UP,DIRECTION_DOWN],[DIRECTION_UP,DIRECTION_LEFT],[DIRECTION_UP,DIRECTION_LEFT,DIRECTION_RIGHT],[DIRECTION_UP,DIRECTION_RIGHT,DIRECTION_DOWN,DIRECTION_LEFT]]
	field parent:TPath
	field color:int
	field kind:int
	field sides:int[]
	field rotation:int
	function create:TTrace(kind:int,parent:TPath)
		local t:TTrace = new TTrace
		t.sides = tracemaps[kind][..]
		t.parent = parent
		t.kind = kind
		t.rotation = 0
		return t
	endfunction
	method rotate(num:int)
		rotation = wrap_var(rotation + num)
		for local i:int = 0 until sides.length
			sides[i] :+ num
			sides[i] = wrap_var(sides[i])
		next
	endmethod
	method is_connected:int(direction:int)
		for local i:int = eachin sides
			if direction = i then return true
		next
		return false
	endmethod
	method add_color(color:int,side_from:int)
		if self.color & color then return
		self.color = mix_colors(self.color,color)
		for local i:int = eachin sides
			if i = side_from then continue
			local n:TTile = parent.get_neighbour(i)
			if n <> null then n.add_color(color,opposite_direction(i))
		next
	endmethod
	method draw(x:int,y:int)
		local rot:int = getrotation()
		setrotation rotation*90
		setconstantcolor(color)
		drawimage trace_images[kind],x,y
		
		setrotation(rot)
	endmethod
endtype






function opposite_direction:int(direction:int)
	select direction
	case DIRECTION_UP; return DIRECTION_DOWN
	case DIRECTION_RIGHT; return DIRECTION_LEFT
	case DIRECTION_DOWN; return DIRECTION_UP
	case DIRECTION_LEFT; return DIRECTION_RIGHT
	endselect
endfunction





function mix_colors:int(color1:int,color2:int)
	return color1 | color2
endfunction



function wrap_var:int(val:int,low:int=0,high:int=3)
	while val > high
		val :- (high-low+1)
	wend
	while val < low
		val :+ (high-low+1)
	wend
	return val
endfunction



function setconstantcolor(color:int,darken:int=false)
	local r:int,g:int,b:int
	select color
	case COLOR_UNDEFINED r=50;g=50;b=50
	case COLOR_RED r=255;g=0;b=0
	case COLOR_GREEN r=0;g=255;b=0
	case COLOR_BLUE r=0;g=0;b=255
	case COLOR_YELLOW r=255;g=255;b=0
	case COLOR_PURPLE r=255;g=0;b=255
	case COLOR_TURQUOISE r=0;g=255;b=255
	case COLOR_WHITE r=255;g=255;b=255
	endselect
	if darken then
		r :/ 2
		g :/ 2
		b :/ 2
	endif
	setcolor r,g,b
endfunction


function strarr_to_intarr:int[](str:string[])
	local result:int[str.length]
	for local i:int = 0 until str.length
		result[i] = int(str[i])
	next
	return result
endfunction


function intarr_to_strarr:string[](in:int[])
	local result:string[in.length]
	for local i:int = 0 until in.length
		result[i] = string(in[i])
	next
	return result
endfunction


function mouse_in_rect:int(mx:int,my:int,x:int,y:int,w:int,h:int)
	return mx >= x and mx <= x+w and my >= y and my <= y+h
endfunction





const COLOR_UNDEFINED:int = 0
const COLOR_RED:int = 1
const COLOR_GREEN:int = 2
const COLOR_BLUE:int = 4
const COLOR_YELLOW:int = COLOR_RED | COLOR_GREEN
const COLOR_PURPLE:int = COLOR_RED | COLOR_BLUE
const COLOR_TURQUOISE:int = COLOR_BLUE | COLOR_GREEN
const COLOR_WHITE:int = COLOR_RED | COLOR_GREEN | COLOR_BLUE

const DIRECTION_UP:int = 0
const DIRECTION_RIGHT:int = 1
const DIRECTION_DOWN:int = 2
const DIRECTION_LEFT:int = 3

const TRACE_2OP:int = 0
const TRACE_2N:int = 1
const TRACE_3:int = 2
const TRACE_4:int = 3



global img_empty:TImage = loadimage("media/empty.png",0)
global img_light:TImage = loadimage("media/light.png",0)
global img_wire:TImage = loadimage("media/shortwire.png",0)
global img_lamp:TImage[] = [loadimage("media/lamp_off.png"),loadimage("media/lamp_on.png")]
global trace_images:TImage[] = [loadimage("media/2op.png",0),loadimage("media/2n.png",0),loadimage("media/3.png",0),loadimage("media/4.png",0)]
midhandleimage img_empty
midhandleimage img_light
midhandleimage img_wire
for local i:TImage = eachin (img_lamp+trace_images)
	midhandleimage i
next

global gamefield:TGamefield



seedrnd millisecs()

graphics 600,600,0
setblend ALPHABLEND




